# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import BaseHTTPServer
import logging
import threading
import urllib2


log = logging.getLogger(__name__)


class HTTPServer(BaseHTTPServer.HTTPServer):
    # shutdown mechanism borrowed from gocept.selenium.static.HTTPServer

    _continue = True

    @classmethod
    def start(cls, host, port, bender, username, password):
        server_address = (host, port)
        BenderRequestHandler.bender = bender
        BenderRequestHandler.username = username
        BenderRequestHandler.password = password
        httpd = cls(server_address, BenderRequestHandler)
        thread = threading.Thread(target=httpd.serve_until_shutdown)
        thread.daemon = True
        thread.start()
        log.info('HTTP server started on %r', server_address)
        return httpd

    def serve_until_shutdown(self):
        while self._continue:
            self.handle_request()

    def shutdown(self):
        self._continue = False
        # We fire a last request at the server in order to take it out of the
        # while loop in `self.serve_until_shutdown`.
        try:
            urllib2.urlopen(
                'http://%s:%s/die' % (self.server_name, self.server_port),
                timeout=1)
        except urllib2.URLError:
            # If the server is already shut down, we receive a socket error,
            # which we ignore.
            pass
        self.server_close()


class BaseHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    def log_message(self, format, *args):
        pass


class BenderRequestHandler(BaseHTTPRequestHandler):

    bender = None
    username = None
    password = None

    def authorize(self):
        auth_header = self.headers.getheader('Authorization')
        hash = '%s:%s' % (self.username, self.password)
        hash = hash.encode('base64').strip()
        if auth_header == 'Basic %s' % hash:
            return True
        return False

    def do_POST(self):
        if not self.authorize():
            self.send_response(401)
            self.send_header(
                'WWW-Authenticate', 'Basic realm="Bender Jabber Bot"')
            self.render(
                'ERROR - Unauthorized: You sent no or invalid credentials')
            return

        length = int(self.headers['content-length'])
        data = self.rfile.read(length)
        self.bender.say(data)
        self.send_response(200)
        log.info('Received POST from %r', self.client_address[0])
        self.render("OK - I've said that to %s" % self.bender.chatroom)

    def do_GET(self):
        self.send_response(200)
        self.render(
            "OK - Hello human! I'm Bender. If you have something"
            " for me to say, please POST it here.")

    def render(self, text):
        self.send_header('Content-type', 'text/plain; charset=utf-8')
        self.end_headers()
        if not text.endswith('\n'):
            text += '\n'
        self.wfile.write(text.encode('utf-8'))



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import gocept.bender.http
import mock
import unittest
import urllib2


class HTTPServerTest(unittest.TestCase):

    port = 12345

    def setUp(self):
        self.bender = mock.Mock()
        self.httpd = gocept.bender.http.HTTPServer.start(
            'localhost', self.port, self.bender, 'user', 'pass')

    def tearDown(self):
        self.httpd.shutdown()

    def make_request(self, user=None, password=None, data='foo'):
        url = 'http://localhost:%s' % self.port
        passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, url, user, password)
        authhandler = urllib2.HTTPBasicAuthHandler(passman)
        opener = urllib2.build_opener(authhandler)
        return opener.open(url, data)

    def test_no_credentials_should_return_unauthorized(self):
        try:
            self.make_request()
        except urllib2.HTTPError, e:
            self.assertEqual(401, e.getcode())
        else:
            self.fail('nothing raised')

        self.assertFalse(self.bender.say.called)

    def test_wrong_credentials_should_return_unauthorized(self):
        try:
            self.make_request(user='wrong', password='invalid')
        except urllib2.HTTPError, e:
            self.assertEqual(401, e.getcode())
        else:
            self.fail('nothing raised')

        self.assertFalse(self.bender.say.called)

    def test_correct_credentials_should_work(self):
        self.make_request(user='user', password='pass')
        self.bender.say.assert_called_with('foo')



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import Queue
import gocept.bender.http
import gocept.bender.interfaces
import gocept.bender.quote
import jabberbot
import logging
import socket
import sys
import time
import zope.event


log = logging.getLogger(__name__)


class Bender(jabberbot.JabberBot):

    def __init__(self, user, password, chatroom):
        super(Bender, self).__init__(
            user, password, res='bender.' + socket.gethostname())
        self.chatroom = chatroom
        self.join_room(self.chatroom, 'bender')
        log.info('joined %s', self.chatroom)
        self.messages = Queue.Queue()

    def idle_proc(self):
        try:
            message = self.messages.get_nowait()
        except Queue.Empty:
            return super(Bender, self).idle_proc()
        log.info('sending %r to %s', message, self.chatroom)
        self.send(self.chatroom, message, message_type='groupchat')

    def say(self, message):
        self.messages.put(message)

    def callback_message(self, connection, msg):
        # XXX we don't call super at the moment, so botcmd is disabled
        self.__lastping = time.time()
        log.debug('received message %r', (
                msg.getType(), self.get_sender_username(msg),
                msg.getProperties(), msg.getBody()))
        zope.event.notify(
            gocept.bender.interfaces.MessageReceivedEvent(msg))


def main(**kw):
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(
            '%(asctime)s %(levelname)s %(message)s'))
    logging.root.handlers = [handler]
    logging.root.setLevel(getattr(logging, kw['loglevel']))
    bender = Bender(kw['jabber_user'], kw['jabber_password'], kw['chatroom'])
    host, port = kw['http_address'].split(':')
    httpd = gocept.bender.http.HTTPServer.start(
        host, int(port), bender, kw['http_user'], kw['http_password'])
    quote = gocept.bender.quote.QuoteTrigger.start(bender, **kw)
    try:
        bender.serve_forever()
    finally:
        httpd.shutdown()
        quote.stop()



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

from gocept.bender.quote import QuoteTrigger
import datetime
import gocept.bender.bot
import mock
import unittest
import zope.event


class QuoteTriggerTest(unittest.TestCase):

    def test_should_not_talk_twice_within_time_threshold(self):
        bender = mock.Mock()
        trigger = QuoteTrigger(bender)
        trigger.join_wait = 0
        trigger.min_silence_duration = datetime.timedelta(minutes=1)
        trigger.speaking_probability = 1.0
        trigger.min_human_messages = 0
        trigger.maybe_say_something()
        trigger.maybe_say_something()
        self.assertEqual(1, bender.say.call_count)

    def test_should_not_talk_until_humans_have_said_enough(self):
        bender = mock.Mock()
        trigger = QuoteTrigger(bender)
        trigger.join_wait = 0
        trigger.speaking_probability = 1.0
        trigger.min_silence_duration = datetime.timedelta(minutes=0)
        trigger.min_human_messages = 1

        trigger.maybe_say_something()
        self.assertFalse(bender.say.called)

        msg = mock.Mock()
        msg.getType.return_value = 'groupchat'
        msg.getFrom().getResource.return_value = 'some_human'
        zope.event.notify(gocept.bender.interfaces.MessageReceivedEvent(msg))

        trigger.maybe_say_something()
        self.assertTrue(bender.say.called)



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

import datetime
import gocept.bender.bot
import gocept.bender.interfaces
import pkg_resources
import random
import threading
import time
import zope.component
import zope.component.event # enable zope.event / zope.component


class QuoteTrigger(object):

    _continue = True
    min_silence_duration = NotImplemented
    min_human_messages = NotImplemented
    speaking_probability = NotImplemented
    join_wait = 30

    def __init__(self, bender, **kw):
        self.bender = bender
        self.last_spoken = datetime.datetime.min
        self.human_message_count = 0
        self.quotes = pkg_resources.resource_string(
            self.__class__.__module__, 'quote.txt').splitlines()
        self.join_time = time.time()

        zope.component.getSiteManager().registerHandler(
            self.count_human_message)

        for key in [
            'min_silence_duration',
            'speaking_probability',
            'min_human_messages']:
            setattr(self, key, kw.get(key, getattr(self, key)))

    @classmethod
    def start(cls, *args, **kw):
        trigger = cls(*args, **kw)
        thread = threading.Thread(target=trigger.run)
        thread.daemon = True
        thread.start()
        return trigger

    def run(self):
        while self._continue:
            self.maybe_say_something()
            time.sleep(1)

    def stop(self):
        self._continue = False

    @zope.component.adapter(gocept.bender.interfaces.MessageReceivedEvent)
    def count_human_message(self, event):
        # skip backlog messages
        if time.time() - self.join_time < self.join_wait:
            return

        message = event.message
        if message.getType() != 'groupchat':
            return
        if message.getFrom().getResource() == 'bender':
            return
        self.human_message_count += 1

    @property
    def may_speak(self):
        now = datetime.datetime.now()
        return (now > self.last_spoken + self.min_silence_duration
                and self.human_message_count >= self.min_human_messages)

    @property
    def should_speak(self):
        return random.random() < self.speaking_probability

    def maybe_say_something(self):
        if not self.may_speak:
            return
        if self.should_speak:
            self.bender.say(random.choice(self.quotes))
            self.last_spoken = datetime.datetime.now()
            self.human_message_count = 0



# Copyright (c) 2011 gocept gmbh & co. kg
# See also LICENSE.txt

from setuptools import setup, find_packages


setup(
    name='gocept.bender',
    version='1.0',
    author='Wolfgang Schnerring',
    author_email='ws@gocept.com',
    url='http://packages.python.org/gocept.bender',
    description="""A Jabber-Bot that can be told things to say via XML-RPC.
""",
    long_description=(
        open('README.txt').read()
        + '\n\n'
        + open('CHANGES.txt').read()),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    license='ZPL',
    namespace_packages=['gocept'],
    install_requires=[
        'dnspython', # xmpp doesn't declare
        'jabberbot',
        'setuptools',
        'xmpppy', # jabberbot doesn't declare
        'zope.component',
        'zope.event',
    ],
    extras_require=dict(test=[
        'mock',
    ]),
    entry_points=dict(console_scripts=[
        'bender-server = gocept.bender.bot:main',
    ]),
)
